@@ -68,10 +68,33 @@ fn main() {
 
     // Iterate all input elfs. Convert them to Tock friendly binaries and then
     // add them to the TAB file.
-    for elf_path in opt.input {
-        let tbf_path = elf_path.with_extension("tbf");
+    for elf_file in opt.input {
+        let elffile = elf::File::open_path(&elf_file.path).expect("Could not open the .elf file.");
 
-        let elffile = elf::File::open_path(&elf_path).expect("Could not open the .elf file.");
+        // The TBF will be written to the same place as the ELF, with a .tbf
+        // extension.
+        let tbf_path = elf_file.path.with_extension("tbf");
+
+        // Get the name of the architecture for the TBF. This will be used to
+        // name the TBF in the TAB, as the file name is expected to be
+        // `<architecture>.tbf`.
+        let architecture = if let Some(ref architecture) = elf_file.architecture {
+            // The caller of elf2tab explicitly told us the architecture via
+            // command line arguments.
+            architecture.clone()
+        } else {
+            // Otherwise, we must assume that the elf was named as
+            // `<architecture>.elf` and use the base name as the architecture.
+            elf_file
+                .path
+                .file_name()
+                .unwrap()
+                .to_str()
+                .unwrap()
+                .to_string()
+        };
+        // Use the architecture to name the TBF in the TAB.
+        let tab_tbf_name = format!("{}.tbf", architecture);
 
         if opt.output.clone() == tbf_path.clone() {
             panic!(
@@ -113,6 +136,7 @@ fn main() {
             opt.kernel_heap_size,
             opt.protected_region_size,
             opt.permissions.to_vec(),
+            (opt.write_id, opt.read_ids.clone(), opt.access_ids.clone()),
             minimum_tock_kernel_version,
             add_trailing_padding,
         )
@@ -123,8 +147,7 @@ fn main() {
 
         // Add the file to the TAB tar file.
         outfile.seek(io::SeekFrom::Start(0)).unwrap();
-        tab.append_file(tbf_path.file_name().unwrap(), &mut outfile)
-            .unwrap();
+        tab.append_file(tab_tbf_name, &mut outfile).unwrap();
     }
 }
 
@@ -145,11 +168,12 @@ fn elf_to_tbf<W: Write>(
     output: &mut W,
     package_name: Option<String>,
     verbose: bool,
-    stack_len: u32,
+    stack_len: Option<u32>,
     app_heap_len: u32,
     kernel_heap_len: u32,
     protected_region_size_arg: Option<u32>,
     permissions: Vec<(u32, u32)>,
+    storage_ids: (Option<u32>, Option<Vec<u32>>, Option<Vec<u32>>),
     kernel_version: Option<(u16, u16)>,
     trailing_padding: bool,
 ) -> io::Result<()> {
@@ -163,6 +187,20 @@ fn elf_to_tbf<W: Write>(
     }
     sections_sort.sort_by_key(|s| s.1);
 
+    let stack_len = stack_len
+        // not provided, read from binary
+        .or_else(|| {
+            input.sections.iter().find_map(|section| {
+                if section.shdr.name == ".stack" {
+                    Some(section.shdr.size as u32)
+                } else {
+                    None
+                }
+            })
+        })
+        // nothing in binary, use default
+        .unwrap_or(2048);
+
     // Keep track of how much RAM this app will need.
     let mut minimum_ram_size: u32 = 0;
 
@@ -359,6 +397,7 @@ fn elf_to_tbf<W: Write>(
         fixed_address_ram,
         fixed_address_flash,
         permissions,
+        storage_ids,
         kernel_version,
     );
     // If a protected region size was passed, confirm the header will fit.
